#    (c)  Copyright 2022 Fungible, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
from unittest import mock
import uuid

from cinder import context
from cinder import exception
from cinder.image import image_utils
from cinder.objects import fields
from cinder.tests.unit import fake_constants
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit.image import fake as fake_image
from cinder.tests.unit.volume.drivers.fungible import test_adapter
from cinder import volume
from cinder.volume import configuration
from cinder.volume.drivers.fungible import constants
from cinder.volume.drivers.fungible import driver
from cinder.volume.drivers.fungible import rest_client
from cinder.volume.drivers.fungible import \
    swagger_api_client as swagger_client
from cinder.volume import volume_types
from cinder.volume import volume_utils

common_success_res = swagger_client.CommonResponseFields(
    status=True, message='healthy')
common_failure_res = swagger_client.CommonResponseFields(
    status=False, message='error')
success_uuid = swagger_client.ResponseDataWithCreateUuid(
    status=True, data={'uuid': 'mock_id'})
success_response = swagger_client.SuccessResponseFields(
    status=True, message="mock_message")
get_volume_details = swagger_client.ResponseDataWithCreateUuid(
    status=True,
    data={"dpu": "mock_dpu_uuid", "secy_dpu": "mock_dpu_uuid",
          "ports": {"mock_id": {"host_nqn": "mock_nqn",
                    "host_uuid": "mock_host_id", "transport": "TCP"}}})
get_topology = swagger_client.ResponseDpuDriveHierarchy(
    status=True,
    data={"mock_device_uuid": {
          "available": True, "dpus": [{"dataplane_ip": "mock_dataplae_ip",
                                       "uuid": "mock_dpu_uuid"}]}})
get_host_id_list = swagger_client.ResponseDataWithListOfHostUuids(
    status=True,
    data={"total_hosts_with_fac": 0, "total_hosts_without_fac": 1,
          "host_uuids": ["mock_host_id"]})
get_host_info = swagger_client.ResponseDataWithHostInfo(
    status=True,
    data={"host_uuid": "mock_host_id", "host_nqn": "mock_nqn",
          "fac_enabled": False})
fetch_hosts_with_ids = swagger_client.ResponseDataWithListOfHosts(
    status=True,
    data=[
        {
            "host_uuid": "mock_host_id", "host_nqn": "mock_nqn",
            "fac_enabled": False
        }
    ])
create_copy_task = swagger_client.ResponseCreateVolumeCopyTask(
    status=True, data={'task_uuid': 'mock_id'})
get_task_success = swagger_client.ResponseGetVolumeCopyTask(
    status=True, data={'task_state': 'SUCCESS'})


class FungibleDriverTest(unittest.TestCase):
    def setUp(self):
        super(FungibleDriverTest, self).setUp()
        self.configuration = mock.Mock(spec=configuration.Configuration)
        self.configuration.san_ip = '127.0.0.1'
        self.configuration.san_api_port = 443
        self.configuration.san_login = 'admin'
        self.configuration.san_password = 'password'
        self.configuration.nvme_connect_port = 4420
        self.configuration.api_enable_ssl = False
        self.driver = driver.FungibleDriver(configuration=self.configuration)
        self.driver.do_setup(context=None)
        self.context = context.get_admin_context()
        self.api_exception = swagger_client.ApiException(
            status=400,
            reason="Bad Request",
            http_resp=self.get_api_exception_response())

    @staticmethod
    def get_volume():
        volume = fake_volume.fake_volume_obj(mock.MagicMock())
        volume.size = 1
        volume.provider_id = fake_constants.UUID1
        volume.migration_status = ''
        volume.id = str(uuid.uuid4())
        volume.display_name = 'volume'
        volume.host = 'mock_host_name'
        volume.volume_type_id = fake_constants.VOLUME_TYPE_ID
        volume.metadata = {}
        return volume

    @staticmethod
    def get_snapshot():
        snapshot = fake_snapshot.fake_snapshot_obj(mock.MagicMock())
        snapshot.display_name = 'snapshot'
        snapshot.provider_id = fake_constants.UUID1
        snapshot.id = str(uuid.uuid4())
        snapshot.volume = FungibleDriverTest.get_volume()
        return snapshot

    @staticmethod
    def get_connector():
        return {"nqn": "mock_nqn"}

    @staticmethod
    def get_specs():
        return {
            constants.FSC_SPACE_ALLOCATION_POLICY: "write_optimized",
            constants.FSC_COMPRESSION: "true",
            constants.FSC_QOS_BAND: "gold",
            constants.FSC_SNAPSHOTS: "false",
            constants.FSC_BLK_SIZE: "4096"
        }

    @staticmethod
    def get_metadata():
        return {
            constants.FSC_SPACE_ALLOCATION_POLICY: "write_optimized",
            constants.FSC_COMPRESSION: "false",
            constants.FSC_QOS_BAND: "bronze",
            constants.FSC_EC_SCHEME: constants.EC_4_2
        }

    @staticmethod
    def get_api_exception_response():
        return test_adapter.MockResource(
            status=False,
            data='{"error_message":"mock_error_message","status":false}')

    '''@staticmethod
    def get_volume_details():
        return {
            "data": {
                "ports": {
                    "mock_id": {
                        "host_nqn": "mock_nqn",
                        "ip": "127.0.0.1"
                    }
                }
            }
        }'''

    def test_get_driver_options(self):
        self.assertIsNotNone(self.driver.get_driver_options())

    def test_volume_stats(self):
        self.assertIsNotNone(self.driver.get_volume_stats())

    @mock.patch.object(swagger_client.ApigatewayApi, 'get_fc_health')
    def test_check_for_setup_error_success(self, mock_success_response):
        mock_success_response.return_value = common_success_res
        result = self.driver.check_for_setup_error()
        self.assertIsNone(result)

    @mock.patch.object(swagger_client.ApigatewayApi, 'get_fc_health')
    def test_check_for_setup_error_fail(self, mock_staus):
        mock_staus.return_value = common_failure_res
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.check_for_setup_error()

    @mock.patch.object(rest_client.RestClient, 'check_for_setup_error')
    def test_check_for_setup_error_exception(self, mock_staus):
        mock_staus.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.check_for_setup_error()

    @mock.patch.object(rest_client.RestClient, 'check_for_setup_error')
    def test_check_for_setup_error_api_exception(self, mock_exception):
        mock_exception.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.check_for_setup_error()

    @mock.patch.object(volume_types, 'get_volume_type')
    def test_get_volume_stats_without_volume_type(self, mock_get_volume_type):
        volume = self.get_volume()
        volume.volume_type_id = fake_constants.VOLUME_TYPE_ID
        mock_get_volume_type.return_value = {"extra_specs": self.get_specs()}
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver._get_volume_type_extra_specs(self, volume=volume)

    @mock.patch.object(volume_types, 'get_volume_type')
    def test_get_volume_stats_with_volume_type(self, mock_get_volume_type):
        volume = {"volume_type_id": fake_constants.VOLUME_TYPE_ID}
        extra_specs = self.get_specs()
        extra_specs.update({constants.FSC_VOL_TYPE: constants.VOLUME_TYPE_RAW})
        mock_get_volume_type.return_value = {"extra_specs": extra_specs}
        self.assertIsNotNone(
            self.driver._get_volume_type_extra_specs(self, volume=volume))

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_volume(self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_volume.return_value = success_uuid
        ret = self.driver.create_volume(volume)
        self.assertIsNotNone(ret)
        self.assertEqual(volume['size'], ret['size'])

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_ec_volume_8_2(self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{constants.FSC_EC_SCHEME: "8_2"},
                          constants.VOLUME_TYPE_EC])
        mock_create_volume.return_value = success_uuid
        ret = self.driver.create_volume(volume)
        self.assertIsNotNone(ret)
        self.assertEqual(volume['size'], ret['size'])

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_replicated_volume(self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_REPLICA])
        ret = self.driver.create_volume(volume)
        mock_create_volume.return_value = success_uuid
        self.assertIsNotNone(ret)
        self.assertEqual(volume['size'], ret['size'])

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_volume_with_specs(self, mock_create_volume):
        volume = self.get_volume()
        mock_ret = self.get_specs()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[mock_ret, constants.VOLUME_TYPE_EC])
        mock_create_volume.return_value = success_uuid
        ret = self.driver.create_volume(volume)
        self.assertIsNotNone(ret)
        self.assertEqual(volume['size'], ret['size'])

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_volume_with_metadata(self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_volume.return_value = success_uuid
        volume['metadata'].update(self.get_metadata())
        ret = self.driver.create_volume(volume)
        self.assertIsNotNone(ret)
        self.assertEqual(volume['size'], ret['size'])

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_volume_with_fault_domains(self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_RAW])
        volume['metadata'].update(self.get_metadata())
        volume['metadata'].update({constants.FSC_FD_IDS: 'fake_id1, fake_id2'})
        volume['metadata'].update(
            {constants.FSC_FD_OP: constants.FSC_FD_OPS[0]})
        mock_create_volume.return_value = success_uuid
        ret = self.driver.create_volume(volume)
        self.assertIsNotNone(ret)
        self.assertEqual(volume['size'], ret['size'])

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_negative_create_volume_with_fault_domains(
            self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_RAW])
        volume['metadata'].update(self.get_metadata())
        volume['metadata'].update(
            {constants.FSC_FD_IDS: 'fake_id1,fake_id2,fake_id3'})
        volume['metadata'].update({constants.FSC_FD_OP: 'mock_value'})
        mock_create_volume.return_value = success_uuid
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_volume(volume)

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_negative_create_volume_without_fault_domains_op(
            self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_RAW])
        volume['metadata'].update(self.get_metadata())
        volume['metadata'].update({constants.FSC_FD_IDS: 'fake_id1,fake_id2'})
        mock_create_volume.return_value = success_uuid
        ret = self.driver.create_volume(volume)
        self.assertIsNotNone(ret)
        self.assertEqual(volume['size'], ret['size'])

    def test_negative_create_volume_with_metadata(self):
        volume = self.get_volume()
        volume['metadata'].update(self.get_specs())
        volume['metadata'].update({constants.FSC_QOS_BAND: 'wrong value'})
        volume['metadata'].update(
            {constants.FSC_SPACE_ALLOCATION_POLICY: 'wrong value'})
        volume['metadata'].update({constants.FSC_COMPRESSION: 'wrong value'})
        volume['metadata'].update({constants.FSC_EC_SCHEME: 'wrong value'})
        volume['metadata'].update({constants.FSC_SNAPSHOTS: 'wrong value'})
        volume['metadata'].update({constants.FSC_BLK_SIZE: 'wrong value'})
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_volume(volume)

    def test_negative_encrypted_create_volume(self):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        volume['metadata'].update(self.get_metadata())
        volume['metadata'].update({constants.FSC_KMIP_SECRET_KEY: 'fake key'})
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_volume(volume)

    def test_negative_create_volume_with_specs(self):
        volume = self.get_volume()
        mock_ret = self.get_specs()
        mock_ret.update({constants.FSC_QOS_BAND: 'wrong value'})
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[mock_ret, constants.VOLUME_TYPE_EC])
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_volume(volume)

    @mock.patch.object(rest_client.RestClient, 'create_volume')
    def test_negative_create_volume_api_exception(self, mock_create_volume):
        volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_volume.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_volume(volume)

    @mock.patch.object(swagger_client.StorageApi, 'delete_volume')
    def test_delete_volume(self, mock_delete_volume):
        volume = self.get_volume()
        mock_delete_volume.return_value = success_response
        self.assertIsNone(self.driver.delete_volume(volume))

    @mock.patch.object(rest_client.RestClient, 'delete_volume')
    def test_negative_delete_volume_exception(self, mock_delete_volume):
        mock_volume = self.get_volume()
        mock_volume['provider_id'] = fake_constants.UUID1
        mock_delete_volume.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.delete_volume(mock_volume)

    def test_negative_delete_volume_without_provider_id(self):
        volume = self.get_volume()
        volume['provider_id'] = None
        self.assertIsNone(self.driver.delete_volume(volume))

    def test_negative_delete_volume_without_provider_id_attr(self):
        volume = self.get_volume()
        del volume.provider_id
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.delete_volume(volume)

    @mock.patch.object(swagger_client.StorageApi, 'delete_volume')
    def test_negative_delete_volume_api_exception(self, mock_delete_volume):
        volume = self.get_volume()
        mock_delete_volume.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.delete_volume(volume)

    @mock.patch.object(swagger_client.StorageApi, 'get_volume')
    @mock.patch.object(swagger_client.StorageApi, 'attach_volume')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_id_list')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_info')
    @mock.patch.object(swagger_client.TopologyApi, 'get_hierarchical_topology')
    def test_initialize_connection(
            self, mock_get_topology, mock_get_host_info, mock_get_host_id_list,
            mock_attach_volume, mock_get_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_attach_volume.return_value = success_uuid
        mock_get_volume.return_value = get_volume_details
        mock_get_host_id_list.return_value = get_host_id_list
        mock_get_host_info.return_value = get_host_info
        mock_get_topology.return_value = get_topology
        conn_info = self.driver.initialize_connection(mock_volume, connector)
        self.assertIsNotNone(conn_info)
        self.assertEqual(conn_info.get("driver_volume_type"), "nvmeof")
        self.assertIsNotNone(conn_info.get("data"))
        self.assertEqual(
            conn_info.get("data").get("vol_uuid"), fake_constants.UUID1)
        self.assertEqual(conn_info.get("data").get("host_nqn"),
                         self.get_connector().get("nqn"))
        '''Add more validation here'''

    def test_negative_initialize_connection_without_nqn(self):
        mock_volume = self.get_volume()
        connector = {}
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.initialize_connection(mock_volume, connector)

    def test_negative_initialize_connection_without_provider_id(self):
        mock_volume = {}
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_volume["provider_id"] = None
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.initialize_connection(mock_volume, connector)

    @mock.patch.object(swagger_client.StorageApi, 'attach_volume')
    def test_negative_initialize_connection_api_exception(
            self, mock_attach_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_attach_volume.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.initialize_connection(mock_volume, connector)

    @mock.patch.object(swagger_client.StorageApi, 'attach_volume')
    def test_initialize_connection_exception(self, mock_attach_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_attach_volume.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.initialize_connection(mock_volume, connector)

    @mock.patch.object(swagger_client.StorageApi, 'get_volume')
    @mock.patch.object(swagger_client.StorageApi, 'attach_volume')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_id_list')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_info')
    @mock.patch.object(swagger_client.TopologyApi, 'get_hierarchical_topology')
    def test_initialize_connection_iops_connection(
            self, mock_get_topology, mock_get_host_info, mock_get_host_id_list,
            mock_attach_volume, mock_get_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_attach_volume.return_value = success_uuid
        mock_get_volume.return_value = get_volume_details
        mock_get_host_id_list.return_value = get_host_id_list
        mock_get_host_info.return_value = get_host_info
        mock_get_topology.return_value = get_topology
        connector[constants.FSC_IOPS_IMG_MIG] = True
        conn_info = self.driver.initialize_connection(mock_volume, connector)
        self.assertIsNotNone(conn_info)

    @mock.patch.object(swagger_client.StorageApi, 'get_volume')
    @mock.patch.object(swagger_client.StorageApi, 'attach_volume')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_id_list')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_info')
    @mock.patch.object(swagger_client.TopologyApi, 'get_hierarchical_topology')
    def test_initialize_connection_iops_migration(
            self, mock_get_topology, mock_get_host_info, mock_get_host_id_list,
            mock_attach_volume, mock_get_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_attach_volume.return_value = success_uuid
        mock_get_volume.return_value = get_volume_details
        mock_get_host_id_list.return_value = get_host_id_list
        mock_get_host_info.return_value = get_host_info
        mock_get_topology.return_value = get_topology
        mock_volume['migration_status'] = "migrating"
        conn_info = self.driver.initialize_connection(mock_volume, connector)
        self.assertIsNotNone(conn_info)

    @mock.patch.object(swagger_client.StorageApi, 'get_volume')
    @mock.patch.object(swagger_client.StorageApi, 'delete_port')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_id_list')
    @mock.patch.object(swagger_client.TopologyApi, 'fetch_hosts_with_ids')
    def test_terminate_connection(
            self, mock_fetch_hosts_with_ids, mock_get_host_id_list,
            mock_detach_volume, mock_get_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_get_volume.return_value = get_volume_details
        mock_get_host_id_list.return_value = get_host_id_list
        mock_fetch_hosts_with_ids.return_value = fetch_hosts_with_ids
        mock_detach_volume.return_value = success_uuid
        self.assertIsNone(self.driver.terminate_connection(
            mock_volume, connector))

    def test_negative_terminate_connection_without_provider_id(self):
        mock_volume = {}
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_volume["provider_id"] = None
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.terminate_connection(mock_volume, connector)

    @mock.patch.object(swagger_client.StorageApi, 'get_volume')
    @mock.patch.object(swagger_client.StorageApi, 'delete_port')
    @mock.patch.object(swagger_client.TopologyApi, 'get_host_id_list')
    @mock.patch.object(swagger_client.TopologyApi, 'fetch_hosts_with_ids')
    def test_terminate_connection_force_detach(
            self, mock_fetch_hosts_with_ids, mock_get_host_id_list,
            mock_detach_volume, mock_get_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_get_volume.return_value = get_volume_details
        mock_get_host_id_list.return_value = get_host_id_list
        mock_fetch_hosts_with_ids.return_value = fetch_hosts_with_ids
        mock_detach_volume.return_value = success_uuid
        connector = None
        self.assertIsNone(self.driver.terminate_connection(
            mock_volume, connector))

    @mock.patch.object(swagger_client.StorageApi, 'get_volume')
    def test_negative_terminate_connection_without_nqn(self, mock_get_volume):
        mock_volume = self.get_volume()
        mock_get_volume.return_value = get_volume_details
        connector = {}
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.terminate_connection(mock_volume, connector)

    @mock.patch.object(rest_client.RestClient, 'get_volume_detail')
    def test_negative_terminate_connection_without_port(self, mock_output):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_output.return_value = {'data': {'ports': None}}
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.terminate_connection(mock_volume, connector)

    @mock.patch.object(rest_client.RestClient, 'get_volume_detail')
    def test_negative_terminate_connection_with_invalid_port(
            self, mock_output):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_output.return_value = get_volume_details
        connector['nqn'] = "dummy_nqn"
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.terminate_connection(mock_volume, connector)

    @mock.patch.object(swagger_client.StorageApi, 'get_volume')
    @mock.patch.object(swagger_client.StorageApi, 'delete_port')
    def test_negative_terminate_connection_api_exception(
            self, mock_detach_volume, mock_get_volume):
        mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        connector = self.get_connector()
        mock_get_volume.return_value = get_volume_details
        mock_detach_volume.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.terminate_connection(mock_volume, connector)

    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_volume_from_ec_snapshot(self, mock_create_volume):
        mock_snapshot = self.get_snapshot()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_snapshot.volume = self.get_volume()
        mock_snapshot.provider_id = fake_constants.UUID1
        mock_volume2 = self.get_volume()
        mock_create_volume.return_value = success_uuid
        new_vol_ret = self.driver.create_volume_from_snapshot(
            mock_volume2, mock_snapshot)
        self.assertIsNotNone(new_vol_ret)
        self.assertEqual(mock_volume2['size'], new_vol_ret['size'])

    @mock.patch.object(rest_client.RestClient, 'create_volume')
    def test_create_volume_from_snapshot_exception(
            self, mock_get_volume_detail):
        mock_volume = self.get_volume()
        mock_snapshot = self.get_snapshot()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_get_volume_detail.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_volume_from_snapshot(mock_volume, mock_snapshot)

    @mock.patch.object(rest_client.RestClient, 'create_volume')
    def test_create_volume_from_snapshot_APIException(
            self, mock_create_volume):
        mock_volume = self.get_volume()
        mock_snapshot = self.get_snapshot()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_volume.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_volume_from_snapshot(mock_volume, mock_snapshot)

    @mock.patch.object(swagger_client.StorageApi, 'delete_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'delete_snapshot')
    @mock.patch.object(swagger_client.StorageApi, 'get_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'create_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    @mock.patch.object(swagger_client.StorageApi, 'create_snapshot')
    def test_create_cloned_ec_volume(
            self, mock_create_snapshot, mock_create_volume,
            mock_create_volume_copy_task, mock_get_task, mock_delete_snapshot,
            mock_delete_task):
        target_mock_volume = self.get_volume()
        source_mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_snapshot.return_value = success_uuid
        mock_create_volume.return_value = success_uuid
        mock_create_volume_copy_task.return_value = create_copy_task
        mock_get_task.return_value = get_task_success
        mock_delete_snapshot.return_value = success_response
        mock_delete_task.return_value = success_response
        self.assertIsNotNone(self.driver.create_cloned_volume(
            target_mock_volume, source_mock_volume))

    @mock.patch.object(swagger_client.StorageApi, 'delete_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'delete_snapshot')
    @mock.patch.object(swagger_client.StorageApi, 'get_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'create_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    @mock.patch.object(swagger_client.StorageApi, 'create_snapshot')
    def test_create_cloned_ec_volume_delete_task_exception(
            self, mock_create_snapshot, mock_create_volume,
            mock_create_volume_copy_task, mock_get_task, mock_delete_snapshot,
            mock_delete_task):
        target_mock_volume = self.get_volume()
        source_mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_snapshot.return_value = success_uuid
        mock_create_volume.return_value = success_uuid
        mock_create_volume_copy_task.return_value = create_copy_task
        mock_get_task.return_value = get_task_success
        mock_delete_snapshot.return_value = success_response
        mock_delete_task.side_effect = self.api_exception
        self.assertIsNotNone(self.driver.create_cloned_volume(
            target_mock_volume, source_mock_volume))

    @mock.patch.object(swagger_client.StorageApi, 'get_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'create_volume_copy_task')
    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_cloned_ec_volume_get_task_api_exception(
            self, mock_create_volume, mock_create_volume_copy_task,
            mock_get_task):
        target_mock_volume = self.get_volume()
        source_mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_volume.return_value = success_uuid
        mock_create_volume_copy_task.return_value = create_copy_task
        mock_get_task.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_cloned_volume(
                target_mock_volume, source_mock_volume)

    @mock.patch.object(rest_client.RestClient, 'copy_volume')
    @mock.patch.object(swagger_client.StorageApi, 'create_volume')
    def test_create_cloned_ec_volume_copy_task_exception(
            self, mock_create_volume, mock_create_volume_copy_task):
        target_mock_volume = self.get_volume()
        source_mock_volume = self.get_volume()
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        mock_create_volume.return_value = success_uuid
        mock_create_volume_copy_task.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_cloned_volume(
                target_mock_volume, source_mock_volume)

    def test_create_clone_in_use_volume(self):
        target_mock_volume = self.get_volume()
        source_mock_volume = self.get_volume()
        source_mock_volume['attach_status'] = "attached"
        self.driver._get_volume_type_extra_specs = mock.Mock(
            return_value=[{}, constants.VOLUME_TYPE_EC])
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_cloned_volume(
                target_mock_volume, source_mock_volume)

    @mock.patch.object(swagger_client.StorageApi, 'create_snapshot')
    def test_create_snapshot(self, mock_create_snapshot):
        mock_volume = self.get_volume()
        snapshot = self.get_snapshot()
        snapshot['volume'] = mock_volume
        mock_create_snapshot.return_value = success_uuid
        ret = self.driver.create_snapshot(snapshot)
        self.assertIsNotNone(ret)

    def test_negative_create_snapshot_without_provider_id(self):
        mock_volume = self.get_volume()
        snapshot = self.get_snapshot()
        snapshot['volume'] = mock_volume
        snapshot['volume']['provider_id'] = None
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_snapshot(snapshot)

    def test_negative_create_snapshot_without_provider_id_attr(self):
        mock_volume = self.get_volume()
        snapshot = self.get_snapshot()
        snapshot.volume = mock_volume
        del snapshot.volume.provider_id
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_snapshot(snapshot)

    @mock.patch.object(rest_client.RestClient, 'create_snapshot')
    def test_negative_create_snapshot_exception(self, mock_create_snapshot):
        snapshot = self.get_snapshot()
        mock_volume = self.get_volume()
        mock_volume['provider_id'] = fake_constants.UUID1
        snapshot['volume'] = mock_volume
        mock_create_snapshot.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_snapshot(snapshot)

    @mock.patch.object(rest_client.RestClient, 'create_snapshot')
    def test_negative_create_snapshot_api_exception(
            self, mock_create_snapshot):
        mock_volume = self.get_volume()
        snapshot = self.get_snapshot()
        snapshot.volume = mock_volume
        mock_create_snapshot.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.create_snapshot(snapshot)

    @mock.patch.object(swagger_client.StorageApi, 'delete_snapshot')
    def test_delete_snapshot(self, mock_delete_snapshot):
        mock_volume = self.get_volume()
        snapshot = self.get_snapshot()
        snapshot['volume'] = mock_volume
        mock_delete_snapshot.return_value = success_response
        self.assertIsNone(self.driver.delete_snapshot(snapshot))

    def test_negative_delete_snapshot_without_provider_id(self):
        snapshot = self.get_snapshot()
        snapshot['provider_id'] = None
        self.assertIsNone(self.driver.delete_snapshot(snapshot))

    def test_negative_delete_snapshot_without_provider_id_attr(self):
        mock_volume = self.get_volume()
        snapshot = self.get_snapshot()
        snapshot.volume = mock_volume
        del snapshot.provider_id
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.delete_snapshot(snapshot)

    @mock.patch.object(rest_client.RestClient, 'delete_snapshot')
    def test_negative_delete_snapshot_exception(self, mock_delete_snapshot):
        snapshot = self.get_snapshot()
        mock_volume = self.get_volume()
        mock_volume['provider_id'] = fake_constants.UUID1
        snapshot['volume'] = mock_volume
        mock_delete_snapshot.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.delete_snapshot(snapshot)

    @mock.patch.object(rest_client.RestClient, 'delete_snapshot')
    def test_negative_delete_snapshot_api_exception(
            self, mock_delete_snapshot):
        mock_volume = self.get_volume()
        snapshot = self.get_snapshot()
        snapshot['volume'] = mock_volume
        mock_delete_snapshot.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.delete_snapshot(snapshot)

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_extend_ec_volume_success(self, mock_update_volume):
        mock_volume = self.get_volume()
        new_size = 100
        mock_update_volume.return_value = success_response
        ret = self.driver.extend_volume(mock_volume, new_size)
        self.assertIsNone(ret)

    def test_negative_extend_volume_without_provider_id(self):
        mock_volume = self.get_volume()
        new_size = 100
        mock_volume['provider_id'] = None
        self.assertIsNone(self.driver.extend_volume(mock_volume, new_size))

    def test_negative_extend_volume__without_provider_id_attr(self):
        mock_volume = self.get_volume()
        new_size = 100
        del mock_volume.provider_id
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.extend_volume(mock_volume, new_size)

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_extend_volume_exception(self, mock_update_volume):
        mock_volume = self.get_volume()
        new_size = 100
        mock_update_volume.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.extend_volume(mock_volume, new_size)

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_extend_volume_api_exception(self, mock_update_volume):
        mock_volume = self.get_volume()
        new_size = 100
        mock_update_volume.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.extend_volume(mock_volume, new_size)

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_update_migrated_volume_success(self, mock_rename_volume):
        source_mock_volume = self.get_volume()
        destination_mock_volume = self.get_volume()
        source_mock_volume['host'] = "FSC1"
        destination_mock_volume['host'] = "FSC1"
        mock_rename_volume.return_value = success_response
        self.assertIsNotNone(self.driver.update_migrated_volume(
            self.context, source_mock_volume, destination_mock_volume,
            fields.VolumeStatus.AVAILABLE))

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_update_migrated_volume_without_destination_provider_id(
            self, mock_rename_volume):
        source_mock_volume = self.get_volume()
        destination_mock_volume = self.get_volume()
        destination_mock_volume['provider_id'] = None
        mock_rename_volume.side_effect = success_response
        self.assertIsNotNone(self.driver.update_migrated_volume(
            self.context, source_mock_volume, destination_mock_volume,
            fields.VolumeStatus.AVAILABLE))

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_update_migrated_volume_without_source_provider_id(
            self, mock_rename_volume):
        source_mock_volume = self.get_volume()
        destination_mock_volume = self.get_volume()
        source_mock_volume['provider_id'] = None
        mock_rename_volume.return_value = success_response
        self.assertIsNotNone(self.driver.update_migrated_volume(
            self.context, source_mock_volume, destination_mock_volume,
            fields.VolumeStatus.AVAILABLE))

    @mock.patch.object(rest_client.RestClient, 'rename_volume')
    def test_update_migrated_volume_api_exception(self, mock_rename_volume):
        source_mock_volume = self.get_volume()
        destination_mock_volume = self.get_volume()
        mock_rename_volume[0].side_effect = self.api_exception
        mock_rename_volume[1].return_value = success_response
        self.assertIsNotNone(self.driver.update_migrated_volume(
            self.context, source_mock_volume, destination_mock_volume,
            fields.VolumeStatus.AVAILABLE))

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_update_migrated_volume_backend_exception(
            self, mock_rename_volume):
        source_mock_volume = self.get_volume()
        destination_mock_volume = self.get_volume()
        source_mock_volume['provider_id'] = None
        mock_rename_volume.side_effect = self.api_exception
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.update_migrated_volume(
                self.context, source_mock_volume, destination_mock_volume,
                fields.VolumeStatus.AVAILABLE)

    @mock.patch.object(swagger_client.StorageApi, 'update_volume')
    def test_update_migrated_volume_exception(self, mock_rename_volume):
        source_mock_volume = self.get_volume()
        destination_mock_volume = self.get_volume()
        source_mock_volume['provider_id'] = None
        mock_rename_volume.side_effect = Exception("mock exception")
        with self.assertRaises(exception.VolumeBackendAPIException):
            self.driver.update_migrated_volume(
                self.context, source_mock_volume, destination_mock_volume,
                fields.VolumeStatus.AVAILABLE)

    @mock.patch.object(volume.driver.BaseVD, '_detach_volume')
    @mock.patch.object(image_utils, 'upload_volume')
    @mock.patch.object(volume.driver.BaseVD, '_attach_volume')
    @mock.patch.object(volume_utils, 'brick_get_connector_properties')
    def test_copy_volume_to_image(
            self, mock_get_connector, mock_attach_volume,
            mock_upload_volume, mock_detach):
        mock_volume = self.get_volume()
        image_service = fake_image.FakeImageService()
        self.configuration.use_multipath_for_image_xfer = False
        self.configuration.enforce_multipath_for_image_xfer = False
        local_path = 'dev/sda'
        mock_get_connector.return_value = {}
        attach_info = {'device': {'path': local_path},
                       'conn': {'driver_volume_type': 'nvme',
                                'data': {}, }}
        mock_attach_volume.return_value = [attach_info, mock_volume]
        mock_upload_volume.return_value = None
        mock_detach.return_value = None
        self.driver.wait_for_device = mock.Mock(
            return_value=True)
        self.assertIsNone(
            self.driver.copy_volume_to_image(
                self.context, mock_volume, image_service,
                fake_constants.IMAGE_ID))

    @mock.patch.object(volume.driver.BaseVD, '_detach_volume')
    @mock.patch.object(image_utils, 'fetch_to_raw')
    @mock.patch.object(volume.driver.BaseVD, '_attach_volume')
    @mock.patch.object(volume_utils, 'brick_get_connector_properties')
    def test_copy_image_to_volume(
            self, mock_get_connector, mock_attach_volume,
            mock_fetch_to_raw, mock_detach):
        mock_volume = self.get_volume()
        image_service = fake_image.FakeImageService()
        self.configuration.use_multipath_for_image_xfer = False
        self.configuration.enforce_multipath_for_image_xfer = False
        self.configuration.volume_dd_blocksize = 8
        local_path = 'dev/sda'
        mock_get_connector.return_value = {}
        attach_info = {'device': {'path': local_path},
                       'conn': {'driver_volume_type': 'nvme',
                                'data': {}, }}
        mock_attach_volume.return_value = [attach_info, mock_volume]
        mock_fetch_to_raw.return_value = None
        mock_detach.return_value = None
        self.driver.wait_for_device = mock.Mock(
            return_value=True)
        self.assertIsNone(
            self.driver.copy_image_to_volume(
                self.context, mock_volume, image_service,
                fake_constants.IMAGE_ID))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
from unittest import mock

from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_data as tpd)
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_fake_objects as tpfo)
from cinder.volume.drivers.dell_emc.powermax import nvme_tcp
from cinder.volume.drivers.dell_emc.powermax import rest
from cinder.volume import volume_utils


class PowerMaxNVMeTCPTest(test.TestCase):
    def setUp(self):
        self.data = tpd.PowerMaxData()
        super(PowerMaxNVMeTCPTest, self).setUp()
        self.mock_object(volume_utils, 'get_max_over_subscription_ratio')
        self.configuration = tpfo.FakeConfiguration(
            None, 'NVMeTests', 1, 1, san_ip='1.1.1.1', san_login='smc',
            powermax_array=self.data.array, powermax_srp='SRP_1',
            san_password='smc', san_api_port=8443,
            powermax_port_groups=[self.data.port_group_name_i])
        self.mock_object(rest.PowerMaxRest, '_establish_rest_session',
                         return_value=tpfo.FakeRequestsSession())
        driver = (nvme_tcp.
                  PowerMaxNVMETCPDriver(configuration=self.configuration))
        self.driver = driver
        self.common = self.driver.common
        self.masking = self.common.masking
        self.utils = self.common.utils
        self.rest = self.common.rest
        self.mock_object(
            self.utils, 'get_volumetype_extra_specs',
            return_value=deepcopy(self.data.vol_type_extra_specs))

    def test_create_volume(self):
        with mock.patch.object(self.common, 'create_volume') as mock_create:
            self.driver.create_volume(self.data.test_volume)
            mock_create.assert_called_once_with(self.data.test_volume)

    def test_create_volume_from_snapshot(self):
        volume = self.data.test_clone_volume
        snapshot = self.data.test_snapshot
        with mock.patch.object(
                self.common, 'create_volume_from_snapshot') as mock_create:
            self.driver.create_volume_from_snapshot(volume, snapshot)
            mock_create.assert_called_once_with(volume, snapshot)

    def test_create_cloned_volume(self):
        volume = self.data.test_clone_volume
        src_volume = self.data.test_volume
        with mock.patch.object(
                self.common, 'create_cloned_volume') as mock_create:
            self.driver.create_cloned_volume(volume, src_volume)
            mock_create.assert_called_once_with(volume, src_volume)

    def test_delete_volume(self):
        with mock.patch.object(self.common, 'delete_volume') as mock_delete:
            self.driver.delete_volume(self.data.test_volume)
            mock_delete.assert_called_once_with(self.data.test_volume)

    def test_create_snapshot(self):
        with mock.patch.object(self.common, 'create_snapshot') as mock_create:
            self.driver.create_snapshot(self.data.test_snapshot)
            mock_create.assert_called_once_with(
                self.data.test_snapshot, self.data.test_snapshot.volume)

    def test_delete_snapshot(self):
        with mock.patch.object(self.common, 'delete_snapshot') as mock_delete:
            self.driver.delete_snapshot(self.data.test_snapshot)
            mock_delete.assert_called_once_with(
                self.data.test_snapshot, self.data.test_snapshot.volume)

    def test_update_volume_stats(self):
        with mock.patch.object(self.common, 'update_volume_stats',
                               return_value={}) as mock_update:
            self.driver._update_volume_stats()
            mock_update.assert_called_once_with()

    def test_check_for_setup_error_with_valid_versions(self):
        with mock.patch.object(
                self.common.rest, 'get_uni_version',
                return_value=('10.1.0.6', '101')) as mock_uni_version:
            with mock.patch.object(
                    self.common.rest, 'get_vmax_model',
                    return_value=('Powermax_2500')) as mock_pmax_version:
                self.driver.check_for_setup_error()
                mock_uni_version.assert_called_once()
                mock_pmax_version.assert_called_once()

    def test_check_for_setup_error_with_valid_powermax_850_version(self):
        with mock.patch.object(
                self.common.rest, 'get_uni_version',
                return_value=('10.1.0.6', '101')) as mock_uni_version:
            with mock.patch.object(
                    self.common.rest, 'get_vmax_model',
                    return_value=('Powermax_8500')) as mock_pmax_version:
                self.driver.check_for_setup_error()
                mock_uni_version.assert_called_once()
                mock_pmax_version.assert_called_once()

    def test_check_for_setup_error_exception(self):
        with mock.patch.object(
                self.common.rest, 'get_uni_version',
                return_value=('9.2.0.0', '92')) as mock_rest:
            with mock.patch.object(
                    self.common.rest, 'get_vmax_model',
                    return_value=('Powermax_2000')) as mock_pmax_version:
                mock_rest.assert_not_called()
                mock_pmax_version.assert_not_called()
                self.assertRaises(
                    exception.InvalidConfigurationValue,
                    self.driver.check_for_setup_error)

    def test_check_for_setup_error_exception_with_invalid_powermax_version(
            self):
        with mock.patch.object(
                self.common.rest, 'get_uni_version',
                return_value=('10.1.0.6', '101')) as mock_uni_version:
            with mock.patch.object(
                    self.common.rest, 'get_vmax_model',
                    return_value=('Powermax_2000')) as mock_pmax_version:
                mock_uni_version.assert_not_called()
                mock_pmax_version.assert_not_called()
                self.assertRaises(
                    exception.InvalidConfigurationValue,
                    self.driver.check_for_setup_error)

    def test_check_for_setup_error_exception_with_invalid_powermax_version_2(
            self):
        with mock.patch.object(
                self.common.rest, 'get_uni_version',
                return_value=('10.1.0.6', '101')) as mock_uni_version:
            with mock.patch.object(
                    self.common.rest, 'get_vmax_model',
                    return_value=('Powermax_8000')) as mock_pmax_version:
                mock_uni_version.assert_not_called()
                mock_pmax_version.assert_not_called()
                self.assertRaises(
                    exception.InvalidConfigurationValue,
                    self.driver.check_for_setup_error)

    def test_check_for_setup_error_exception_without_unisphere_version(self):
        with mock.patch.object(
                self.common.rest, 'get_uni_version',
                return_value=(None, None)) as mock_rest:
            with mock.patch.object(
                    self.common.rest, 'get_vmax_model',
                    return_value=('Powermax_8500')) as mock_pmax_version:
                mock_rest.assert_not_called()
                mock_pmax_version.assert_not_called()
                self.assertRaises(
                    exception.InvalidConfigurationValue,
                    self.driver.check_for_setup_error)

    def test_check_for_setup_error_exception_without_powermax_version(self):
        with mock.patch.object(
                self.common.rest, 'get_uni_version',
                return_value=('10.1.0.6', '101')) as mock_rest:
            with mock.patch.object(
                    self.common.rest, 'get_vmax_model',
                    return_value=None) as mock_pmax_version:
                mock_rest.assert_not_called()
                mock_pmax_version.assert_not_called()
                self.assertRaises(
                    exception.InvalidConfigurationValue,
                    self.driver.check_for_setup_error)

    def test_ensure_export(self):
        self.driver.ensure_export('context', 'volume')

    def test_create_export(self):
        self.driver.create_export('context', 'volume', 'connector')

    def test_remove_export(self):
        self.driver.remove_export('context', 'volume')

    def test_check_for_export(self):
        self.driver.check_for_export('context', 'volume_id')

    def test_extend_volume(self):
        with mock.patch.object(self.common, 'extend_volume') as mock_extend:
            self.driver.extend_volume(self.data.test_volume, '8')
            mock_extend.assert_called_once_with(self.data.test_volume, '8')

    def test_manage_existing(self):
        with mock.patch.object(self.common, 'manage_existing',
                               return_value={}) as mock_manage:
            external_ref = {u'source-name': u'00002'}
            self.driver.manage_existing(self.data.test_volume, external_ref)
            mock_manage.assert_called_once_with(
                self.data.test_volume, external_ref)

    def test_manage_existing_get_size(self):
        with mock.patch.object(self.common, 'manage_existing_get_size',
                               return_value='1') as mock_manage:
            external_ref = {u'source-name': u'00002'}
            self.driver.manage_existing_get_size(
                self.data.test_volume, external_ref)
            mock_manage.assert_called_once_with(
                self.data.test_volume, external_ref)

    def test_unmanage_volume(self):
        with mock.patch.object(self.common, 'unmanage',
                               return_value={}) as mock_unmanage:
            self.driver.unmanage(self.data.test_volume)
            mock_unmanage.assert_called_once_with(
                self.data.test_volume)

    def test_retype(self):
        host = {'host': self.data.new_host}
        new_type = {'extra_specs': {}}
        with mock.patch.object(self.common, 'retype',
                               return_value=True) as mck_retype:
            self.driver.retype({}, self.data.test_volume, new_type, '', host)
            mck_retype.assert_called_once_with(
                self.data.test_volume, new_type, host)

    def test_failover_host(self):
        with mock.patch.object(
                self.common, 'failover',
                return_value=(self.data.remote_array, [], [])) as mock_fo:
            self.driver.failover_host(self.data.ctx, [self.data.test_volume])
            mock_fo.assert_called_once_with([self.data.test_volume], None,
                                            None)

    def test_enable_replication(self):
        with mock.patch.object(
                self.common, 'enable_replication') as mock_er:
            self.driver.enable_replication(
                self.data.ctx, self.data.test_group, [self.data.test_volume])
            mock_er.assert_called_once()

    def test_disable_replication(self):
        with mock.patch.object(
                self.common, 'disable_replication') as mock_dr:
            self.driver.disable_replication(
                self.data.ctx, self.data.test_group, [self.data.test_volume])
            mock_dr.assert_called_once()

    def test_failover_replication(self):
        with mock.patch.object(
                self.common, 'failover_replication') as mock_fo:
            self.driver.failover_replication(
                self.data.ctx, self.data.test_group, [self.data.test_volume])
            mock_fo.assert_called_once()

    def test_initialize_connection(self):
        with (mock.patch.object(
                self.common, 'initialize_connection',
                return_value=self.data.nvme_tcp_device_info)
                as mock_initialize):
            with mock.patch.object(
                    self.driver, '_populate_data') as mock_populate:
                self.driver.initialize_connection(
                    self.data.test_volume, self.data.connector)
                mock_initialize.assert_called_once_with(
                    self.data.test_volume, self.data.connector)
                mock_populate.assert_called_once_with(
                    self.data.nvme_tcp_device_info)

    def test_terminate_connection(self):
        with mock.patch.object(
                self.common, 'terminate_connection') as mock_terminate:
            self.driver.terminate_connection(
                self.data.test_volume, self.data.connector)
            mock_terminate.assert_called_once_with(
                self.data.test_volume, self.data.connector)

    def test_manage_existing_snapshot(self):
        with mock.patch.object(self.common, 'manage_existing_snapshot',
                               return_value={}) as mock_manage:
            external_ref = {u'source-name': u'00002'}
            self.driver.manage_existing_snapshot(
                self.data.test_snapshot_manage,
                external_ref)
            mock_manage.assert_called_once_with(
                self.data.test_snapshot_manage,
                external_ref)

    def test_manage_existing_snapshot_get_size(self):
        with mock.patch.object(self.common,
                               'manage_existing_snapshot_get_size',
                               return_value='1') as mock_manage:
            external_ref = {u'source-name': u'00002'}
            self.driver.manage_existing_snapshot_get_size(
                self.data.test_snapshot_manage, external_ref)
            mock_manage.assert_called_once_with(
                self.data.test_snapshot_manage)

    def test_unmanage_snapshot(self):
        with mock.patch.object(self.common, 'unmanage_snapshot',
                               return_value={}) as mock_unmanage:
            self.driver.unmanage_snapshot(self.data.test_snapshot_manage)
            mock_unmanage.assert_called_once_with(
                self.data.test_snapshot_manage)

    def test_get_manageable_volumes(self):
        cinder_volumes = marker = limit = offset = sort_keys = sort_dirs = None
        with mock.patch.object(self.common, 'get_manageable_volumes',
                               return_value={}) as mock_manage:
            self.driver.get_manageable_volumes(cinder_volumes,
                                               marker, limit,
                                               offset, sort_keys,
                                               sort_dirs)
            mock_manage.assert_called_once_with(
                marker, limit,
                offset, sort_keys,
                sort_dirs)

    def test_get_manageable_snapshots(self):
        cinder_snapshots = marker = limit = offset = \
            sort_keys = sort_dirs = None
        with mock.patch.object(self.common, 'get_manageable_snapshots',
                               return_value={}) as mock_manage:
            self.driver.get_manageable_snapshots(cinder_snapshots,
                                                 marker, limit,
                                                 offset, sort_keys,
                                                 sort_dirs)
            mock_manage.assert_called_once_with(
                marker, limit,
                offset, sort_keys,
                sort_dirs)

    def test_create_group(self):
        context = {'dummy_key': 'dummy_value'}
        group = 'dummy_group'
        with mock.patch.object(self.common, 'create_group',
                               return_value={}) as mock_group:
            self.driver.create_group(context, group)
            mock_group.assert_called_once_with(
                context, group)

    def test_delete_group(self):
        context = {'dummy_key': 'dummy_value'}
        group = 'dummy_group'
        volumes = ['dummy_volume']
        with mock.patch.object(self.common, 'delete_group',
                               return_value={}) as mock_group:
            self.driver.delete_group(context, group, volumes)
            mock_group.assert_called_once_with(
                context, group, volumes)

    def test_create_group_snapshot(self):
        context = {'dummy_key': 'dummy_value'}
        group_snapshot = 'dummy_group_snapshot'
        snapshots = ['dummy_snapshot']
        with mock.patch.object(self.common, 'create_group_snapshot',
                               return_value={}) as mock_group:
            self.driver.create_group_snapshot(context,
                                              group_snapshot,
                                              snapshots)
            mock_group.assert_called_once_with(
                context, group_snapshot, snapshots)

    def test_delete_group_snapshot(self):
        context = {'dummy_key': 'dummy_value'}
        group_snapshot = 'dummy_group_snapshot'
        snapshots = ['dummy_snapshot']
        with mock.patch.object(self.common, 'delete_group_snapshot',
                               return_value={}) as mock_group:
            self.driver.delete_group_snapshot(context,
                                              group_snapshot,
                                              snapshots)
            mock_group.assert_called_once_with(
                context, group_snapshot, snapshots)

    def test_update_group(self):
        context = {'dummy_key': 'dummy_value'}
        group = 'dummy_group'
        add_volumes = ['dummy_add_volume']
        remove_volumes = ['dummy_remove_volume']
        with mock.patch.object(self.common, 'update_group',
                               return_value={}) as mock_group:
            self.driver.update_group(context,
                                     group,
                                     add_volumes,
                                     remove_volumes)
            mock_group.assert_called_once_with(
                group, add_volumes, remove_volumes)

    def test_create_group_from_src(self):
        context = {'dummy_key': 'dummy_value'}
        group = 'dummy_group'
        group_snapshot = 'dummy_group_snapshot'
        snapshots = ['dummy_snapshot']
        volumes = ['dummy_add_volume']
        source_group = 'dummy_source_group'
        source_volumes = ['dummy_source_volume']
        with mock.patch.object(self.common, 'create_group_from_src',
                               return_value={}) as mock_group:
            self.driver.create_group_from_src(context, group, volumes,
                                              group_snapshot, snapshots,
                                              source_group, source_volumes)
            mock_group.assert_called_once_with(
                context, group, volumes,
                group_snapshot, snapshots,
                source_group, source_volumes)



# Copyright 2022 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
FAKE the nvmet library if it's not installed.
This must be imported before cinder/volume/targets/nvmet.py and
cinder/privsep/targets/nvmet.py
"""

import sys
from unittest import mock

from cinder import exception

mock_nvmet_lib = mock.Mock(name='nvmet',
                           Root=type('Root', (mock.Mock, ), {}),
                           Subsystem=type('Subsystem', (mock.Mock, ), {}),
                           Port=type('Port', (mock.Mock, ), {}),
                           Namespace=type('Namespace', (mock.Mock, ),
                                          {'MAX_NSID': 8192}),
                           Host=type('Host', (mock.Mock, ), {}),
                           ANAGroup=type('ANAGroup', (mock.Mock, ), {}),
                           Referral=type('Referral', (mock.Mock, ), {}),
                           nvme=mock.Mock(CFSNotFound=exception.NotFound))

sys.modules['nvmet'] = mock_nvmet_lib
reset_mock = mock_nvmet_lib.reset_mock



# Copyright 2014 - 2017, 2019  StorPool
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import itertools
import re
from unittest import mock

import ddt
from os_brick.initiator import storpool_utils
from os_brick.tests.initiator import test_storpool_utils
from oslo_utils import units

from cinder import exception
from cinder.tests.unit import fake_constants
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers import storpool as driver


volume_types = {
    fake_constants.VOLUME_TYPE_ID: {},
    fake_constants.VOLUME_TYPE2_ID: {'storpool_template': 'ssd'},
    fake_constants.VOLUME_TYPE3_ID: {'storpool_template': 'hdd'}
}
volumes = {}
snapshots = {}


def MockExtraSpecs(vtype):
    return volume_types[vtype]


def mock_volume_types(f):
    def _types_inner_inner1(inst, *args, **kwargs):
        @mock.patch('cinder.volume.volume_types.get_volume_type_extra_specs',
                    new=MockExtraSpecs)
        def _types_inner_inner2():
            return f(inst, *args, **kwargs)

        return _types_inner_inner2()

    return _types_inner_inner1


def volumeName(vid):
    return 'os--volume-{id}'.format(id=vid)


def snapshotName(vtype, vid, more=None):
    return 'os--{t}--{m}--snapshot-{id}'.format(
        t=vtype,
        m="none" if more is None else more,
        id=vid
    )


class MockAPI(object):
    def __init__(self, *args):
        self._disks = {}
        for disk_id in [1, 2, 3, 4]:
            self._disks[disk_id] = {
                'id': disk_id,
                'generationLeft': -1,
                'agCount': 14,
                'agFree': 12,
                'agAllocated': 1
            }
        self._disks[3]['generationLeft'] = 42

        self._templates = [{'name': name} for name in ('ssd', 'hdd')]

    def disks_list(self):
        return self._disks

    def snapshot_create(self, vname, snap):
        snapshots[snap['name']] = dict(volumes[vname])

    def snapshot_update(self, snap, data):
        sdata = snapshots[snap]
        sdata.update(data)

    def snapshot_delete(self, name):
        del snapshots[name]

    def volume_create(self, vol):
        name = vol['name']
        if name in volumes:
            raise storpool_utils.StorPoolAPIError(
                'none',
                {'error': {
                    'descr': 'volume already exists'}})
        data = dict(vol)

        if 'parent' in vol and 'template' not in vol:
            sdata = snapshots[vol['parent']]
            if 'template' in sdata:
                data['template'] = sdata['template']

        if 'baseOn' in vol and 'template' not in vol:
            vdata = volumes[vol['baseOn']]
            if 'template' in vdata:
                data['template'] = vdata['template']

        if 'template' not in data:
            data['template'] = None

        volumes[name] = data

    def volume_delete(self, name):
        del volumes[name]

    def volumes_list(self):
        the_volumes = []
        for volume in volumes:
            the_volumes.append({'name': volume})
        return the_volumes

    def volume_templates_list(self):
        return self._templates

    def volumes_reassign(self, json):
        pass

    def volume_update(self, name, data):
        if 'size' in data:
            volumes[name]['size'] = data['size']

        if 'rename' in data and data['rename'] != name:
            new_name = data['rename']
            volumes[new_name] = volumes[name]
            if volumes[new_name]['name'] == name:
                volumes[new_name]['name'] = new_name
            del volumes[name]

    def volume_revert(self, name, data):
        if name not in volumes:
            raise storpool_utils.StorPoolAPIError(
                'none',
                {'error': {
                    'descr': 'No such volume {name}'.format(name=name)}})

        snapname = data['toSnapshot']
        if snapname not in snapshots:
            raise storpool_utils.StorPoolAPIError(
                'none',
                {'error': {
                    'descr': 'No such snapshot {name}'.format(name=snapname)}})

        volumes[name] = dict(snapshots[snapname])


class MockVolumeDB(object):
    """Simulate a Cinder database with a volume_get() method."""

    def __init__(self, vol_types=None):
        """Store the specified volume types mapping if necessary."""
        self.vol_types = vol_types if vol_types is not None else {}

    def volume_get(self, _context, vid):
        """Get a volume-like structure, only the fields we care about."""
        # Still, try to at least make sure we know about that volume
        return {
            'id': vid,
            'size': volumes[volumeName(vid)]['size'],
            'volume_type': self.vol_types.get(vid),
        }


def MockSPConfig(section = 's01'):
    res = {}
    m = re.match('^s0*([A-Za-z0-9]+)$', section)
    if m:
        res['SP_OURID'] = m.group(1)
    return res


@ddt.ddt
@mock.patch('os_brick.initiator.storpool_utils.get_conf', MockSPConfig)
class StorPoolTestCase(test.TestCase):

    def setUp(self):
        super(StorPoolTestCase, self).setUp()

        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.volume_backend_name = 'storpool_test'
        self.cfg.storpool_template = None
        self.cfg.storpool_replication = 3

        mock_exec = mock.Mock()
        mock_exec.return_value = ('', '')

        self.driver = driver.StorPoolDriver(execute=mock_exec,
                                            configuration=self.cfg)

        with (
            mock.patch(
                'os_brick.initiator.storpool_utils.get_conf'
            ) as get_conf,
            mock.patch(
                'os_brick.initiator.storpool_utils.StorPoolAPI', MockAPI)
        ):
            get_conf.return_value = test_storpool_utils.SP_CONF
            self.driver.check_for_setup_error()

    @ddt.data(
        (5, TypeError),
        ({'no-host': None}, KeyError),
        ({'host': 'sbad'}, driver.StorPoolConfigurationInvalid),
        ({'host': 's01'}, None),
        ({'host': 'none'}, None),
    )
    @ddt.unpack
    def test_validate_connector(self, conn, exc):
        if exc is None:
            self.assertTrue(self.driver.validate_connector(conn))
        else:
            self.assertRaises(exc,
                              self.driver.validate_connector,
                              conn)

    @ddt.data(
        (5, TypeError),
        ({'no-host': None}, KeyError),
        ({'host': 'sbad'}, driver.StorPoolConfigurationInvalid),
    )
    @ddt.unpack
    def test_initialize_connection_bad(self, conn, exc):
        self.assertRaises(exc,
                          self.driver.initialize_connection,
                          None, conn)

    @ddt.data(
        (1, '42', 's01'),
        (2, '616', 's02'),
        (65, '1610', 'none'),
    )
    @ddt.unpack
    def test_initialize_connection_good(self, cid, hid, name):
        c = self.driver.initialize_connection({'id': hid}, {'host': name})
        self.assertEqual('storpool', c['driver_volume_type'])
        self.assertDictEqual({'client_id': cid, 'volume': hid,
                              'access_mode': 'rw'},
                             c['data'])

    def test_noop_functions(self):
        self.driver.terminate_connection(None, None)
        self.driver.create_export(None, None, {})
        self.driver.remove_export(None, None)

    def test_stats(self):
        stats = self.driver.get_volume_stats(refresh=True)
        self.assertEqual('StorPool', stats['vendor_name'])
        self.assertEqual('storpool', stats['storage_protocol'])
        self.assertListEqual(['default', 'template_hdd', 'template_ssd'],
                             sorted([p['pool_name'] for p in stats['pools']]))
        r = re.compile(r'^template_([A-Za-z0-9_]+)$')
        for pool in stats['pools']:
            self.assertEqual(21, pool['total_capacity_gb'])
            self.assertEqual(5, int(pool['free_capacity_gb']))

            self.assertTrue(pool['multiattach'])
            self.assertFalse(pool['QoS_support'])
            self.assertFalse(pool['thick_provisioning_support'])
            self.assertTrue(pool['thin_provisioning_support'])

            if pool['pool_name'] != 'default':
                m = r.match(pool['pool_name'])
                self.assertIsNotNone(m)
                self.assertIsNotNone(m.group(1))
                self.assertEqual(m.group(1), pool['storpool_template'])

    def assertVolumeNames(self, names):
        self.assertListEqual(sorted([volumeName(n) for n in names]),
                             sorted(volumes.keys()))
        self.assertListEqual(sorted([volumeName(n) for n in names]),
                             sorted(data['name'] for data in volumes.values()))

    def assertSnapshotNames(self, specs):
        self.assertListEqual(
            sorted(snapshotName(spec[0], spec[1]) for spec in specs),
            sorted(snapshots.keys()))

    @mock_volume_types
    def test_create_delete_volume(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        self.driver.create_volume({'id': '1', 'name': 'v1', 'size': 1,
                                   'volume_type':
                                       {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertCountEqual([volumeName('1')], volumes.keys())
        self.assertVolumeNames(('1',))
        v = volumes[volumeName('1')]
        self.assertEqual(1 * units.Gi, v['size'])
        self.assertIsNone(v['template'])
        self.assertEqual(3, v['replication'])

        caught = False
        try:
            self.driver.create_volume(
                {'id': '1', 'name': 'v1', 'size': 0,
                 'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        except exception.VolumeBackendAPIException:
            caught = True
        self.assertTrue(caught)

        self.driver.delete_volume({'id': '1'})
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)

        self.driver.create_volume(
            {'id': '1', 'name': 'v1', 'size': 2,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames(('1',))
        v = volumes[volumeName('1')]
        self.assertEqual(2 * units.Gi, v['size'])
        self.assertIsNone(v['template'])
        self.assertEqual(3, v['replication'])

        self.driver.create_volume({'id': '2', 'name': 'v2', 'size': 3,
                                   'volume_type':
                                       {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames(('1', '2'))
        v = volumes[volumeName('2')]
        self.assertEqual(3 * units.Gi, v['size'])
        self.assertIsNone(v['template'])
        self.assertEqual(3, v['replication'])

        self.driver.create_volume(
            {'id': '3', 'name': 'v2', 'size': 4,
             'volume_type': {'id': fake_constants.VOLUME_TYPE2_ID}})
        self.assertVolumeNames(('1', '2', '3'))
        v = volumes[volumeName('3')]
        self.assertEqual(4 * units.Gi, v['size'])
        self.assertEqual('ssd', v['template'])
        self.assertNotIn('replication', v.keys())

        self.driver.create_volume(
            {'id': '4', 'name': 'v2', 'size': 5,
             'volume_type': {'id': fake_constants.VOLUME_TYPE3_ID}})
        self.assertVolumeNames(('1', '2', '3', '4'))
        v = volumes[volumeName('4')]
        self.assertEqual(5 * units.Gi, v['size'])
        self.assertEqual('hdd', v['template'])
        self.assertNotIn('replication', v.keys())

        # Make sure the dictionary is not corrupted somehow...
        v = volumes[volumeName('1')]
        self.assertEqual(2 * units.Gi, v['size'])
        self.assertIsNone(v['template'])
        self.assertEqual(3, v['replication'])

        for vid in ('1', '2', '3', '4'):
            self.driver.delete_volume({'id': vid})
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @mock_volume_types
    def test_update_migrated_volume(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        # Create two volumes
        self.driver.create_volume(
            {'id': '1', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.driver.create_volume(
            {'id': '2', 'name': 'v2', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertCountEqual([volumeName('1'), volumeName('2')],
                              volumes.keys())
        self.assertVolumeNames(('1', '2',))

        # Failure: the "migrated" volume does not even exist
        res = self.driver.update_migrated_volume(None, {'id': '1'},
                                                 {'id': '3', '_name_id': '1'},
                                                 'available')
        self.assertDictEqual({'_name_id': '1'}, res)

        # Success: rename the migrated volume to match the original
        res = self.driver.update_migrated_volume(None, {'id': '3'},
                                                 {'id': '2', '_name_id': '3'},
                                                 'available')
        self.assertDictEqual({'_name_id': None}, res)
        self.assertCountEqual([volumeName('1'), volumeName('3')],
                              volumes.keys())
        self.assertVolumeNames(('1', '3',))

        # Success: swap volume names with an existing volume
        res = self.driver.update_migrated_volume(None, {'id': '1'},
                                                 {'id': '3', '_name_id': '1'},
                                                 'available')
        self.assertDictEqual({'_name_id': None}, res)
        self.assertCountEqual([volumeName('1'), volumeName('3')],
                              volumes.keys())
        self.assertVolumeNames(('1', '3',))

        for vid in ('1', '3'):
            self.driver.delete_volume({'id': vid})
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @mock_volume_types
    def test_clone_extend_volume(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        self.driver.create_volume(
            {'id': '1', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames(('1',))
        self.driver.extend_volume({'id': '1'}, 2)
        self.assertEqual(2 * units.Gi, volumes[volumeName('1')]['size'])

        with mock.patch.object(self.driver, 'db', new=MockVolumeDB()):
            self.driver.create_cloned_volume(
                {
                    'id': '2',
                    'name': 'clo',
                    'size': 3,
                    'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}
                },
                {'id': 1})
        self.assertVolumeNames(('1', '2'))
        self.assertDictEqual({}, snapshots)
        # We do not provide a StorPool template name in either of the volumes'
        # types, so create_cloned_volume() should take the baseOn shortcut.
        vol2 = volumes[volumeName('2')]
        self.assertEqual(vol2['baseOn'], volumeName('1'))
        self.assertNotIn('parent', vol2)

        self.driver.delete_volume({'id': 1})
        self.driver.delete_volume({'id': 2})

        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @ddt.data(*itertools.product(
        [{'id': key} for key in sorted(volume_types.keys())],
        [{'id': key} for key in sorted(volume_types.keys())]))
    @ddt.unpack
    @mock_volume_types
    def test_create_cloned_volume(self, src_type, dst_type):
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        src_template = volume_types[src_type['id']].get('storpool_template')
        dst_template = volume_types[dst_type['id']].get('storpool_template')
        src_name = 's-none' if src_template is None else 's-' + src_template
        dst_name = 'd-none' if dst_template is None else 'd-' + dst_template

        snap_name = snapshotName('clone', '2')

        vdata1 = {
            'id': '1',
            'name': src_name,
            'size': 1,
            'volume_type': src_type,
        }
        self.assertEqual(
            self.driver._template_from_volume(vdata1),
            src_template)
        self.driver.create_volume(vdata1)
        self.assertVolumeNames(('1',))

        vdata2 = {
            'id': 2,
            'name': dst_name,
            'size': 1,
            'volume_type': dst_type,
        }
        self.assertEqual(
            self.driver._template_from_volume(vdata2),
            dst_template)
        with mock.patch.object(self.driver, 'db',
                               new=MockVolumeDB(vol_types={'1': src_type})):
            self.driver.create_cloned_volume(vdata2, {'id': '1'})
        self.assertVolumeNames(('1', '2'))
        vol2 = volumes[volumeName('2')]
        self.assertEqual(vol2['template'], dst_template)

        if src_template == dst_template:
            self.assertEqual(vol2['baseOn'], volumeName('1'))
            self.assertNotIn('parent', vol2)

            self.assertDictEqual({}, snapshots)
        else:
            self.assertNotIn('baseOn', vol2)
            self.assertEqual(vol2['parent'], snap_name)

            self.assertSnapshotNames((('clone', '2'),))
            self.assertEqual(snapshots[snap_name]['template'], dst_template)

        self.driver.delete_volume({'id': '1'})
        self.driver.delete_volume({'id': '2'})
        if src_template != dst_template:
            del snapshots[snap_name]

        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @mock_volume_types
    def test_config_replication(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        save_repl = self.driver.configuration.storpool_replication

        self.driver.configuration.storpool_replication = 3
        stats = self.driver.get_volume_stats(refresh=True)
        pool = stats['pools'][0]
        self.assertEqual(21, pool['total_capacity_gb'])
        self.assertEqual(5, int(pool['free_capacity_gb']))

        self.driver.create_volume(
            {'id': 'cfgrepl1', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames(('cfgrepl1',))
        v = volumes[volumeName('cfgrepl1')]
        self.assertEqual(3, v['replication'])
        self.assertIsNone(v['template'])
        self.driver.delete_volume({'id': 'cfgrepl1'})

        self.driver.configuration.storpool_replication = 2
        stats = self.driver.get_volume_stats(refresh=True)
        pool = stats['pools'][0]
        self.assertEqual(21, pool['total_capacity_gb'])
        self.assertEqual(8, int(pool['free_capacity_gb']))

        self.driver.create_volume(
            {'id': 'cfgrepl2', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames(('cfgrepl2',))
        v = volumes[volumeName('cfgrepl2')]
        self.assertEqual(2, v['replication'])
        self.assertIsNone(v['template'])
        self.driver.delete_volume({'id': 'cfgrepl2'})

        self.driver.create_volume(
            {'id': 'cfgrepl3', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE2_ID}})
        self.assertVolumeNames(('cfgrepl3',))
        v = volumes[volumeName('cfgrepl3')]
        self.assertNotIn('replication', v)
        self.assertEqual('ssd', v['template'])
        self.driver.delete_volume({'id': 'cfgrepl3'})

        self.driver.configuration.storpool_replication = save_repl

        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @mock_volume_types
    def test_config_template(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        save_template = self.driver.configuration.storpool_template

        self.driver.configuration.storpool_template = None

        self.driver.create_volume(
            {'id': 'cfgtempl1', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames(('cfgtempl1',))
        v = volumes[volumeName('cfgtempl1')]
        self.assertEqual(3, v['replication'])
        self.assertIsNone(v['template'])
        self.driver.delete_volume({'id': 'cfgtempl1'})

        self.driver.create_volume(
            {'id': 'cfgtempl2', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE2_ID}})
        self.assertVolumeNames(('cfgtempl2',))
        v = volumes[volumeName('cfgtempl2')]
        self.assertNotIn('replication', v)
        self.assertEqual('ssd', v['template'])
        self.driver.delete_volume({'id': 'cfgtempl2'})

        self.driver.configuration.storpool_template = 'hdd'

        self.driver.create_volume(
            {'id': 'cfgtempl3', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames(('cfgtempl3',))
        v = volumes[volumeName('cfgtempl3')]
        self.assertNotIn('replication', v)
        self.assertEqual('hdd', v['template'])
        self.driver.delete_volume({'id': 'cfgtempl3'})

        self.driver.create_volume(
            {'id': 'cfgtempl4', 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE2_ID}})
        self.assertVolumeNames(('cfgtempl4',))
        v = volumes[volumeName('cfgtempl4')]
        self.assertNotIn('replication', v)
        self.assertEqual('ssd', v['template'])
        self.driver.delete_volume({'id': 'cfgtempl4'})

        self.driver.configuration.storpool_template = save_template

        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @ddt.data(
        # No volume type at all: 'default'
        ('default', None),
        # No storpool_template in the type extra specs: 'default'
        ('default', {'id': fake_constants.VOLUME_TYPE_ID}),
        # An actual template specified: 'template_*'
        ('template_ssd', {'id': fake_constants.VOLUME_TYPE2_ID}),
        ('template_hdd', {'id': fake_constants.VOLUME_TYPE3_ID}),
    )
    @ddt.unpack
    @mock_volume_types
    def test_get_pool(self, pool, volume_type):
        self.assertEqual(pool,
                         self.driver.get_pool({
                             'volume_type': volume_type
                         }))

    @mock_volume_types
    def test_volume_revert(self):
        vol_id = 'rev1'
        vol_name = volumeName(vol_id)
        snap_id = 'rev-s1'
        snap_name = snapshotName('snap', snap_id)

        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        self.driver.create_volume(
            {'id': vol_id, 'name': 'v1', 'size': 1,
             'volume_type': {'id': fake_constants.VOLUME_TYPE_ID}})
        self.assertVolumeNames((vol_id,))
        self.assertDictEqual({}, snapshots)

        self.driver.create_snapshot({'id': snap_id, 'volume_id': vol_id})
        self.assertVolumeNames((vol_id,))
        self.assertListEqual([snap_name], sorted(snapshots.keys()))
        self.assertDictEqual(volumes[vol_name], snapshots[snap_name])
        self.assertIsNot(volumes[vol_name], snapshots[snap_name])

        self.driver.extend_volume({'id': vol_id}, 2)
        self.assertVolumeNames((vol_id,))
        self.assertNotEqual(volumes[vol_name], snapshots[snap_name])

        self.driver.revert_to_snapshot(None, {'id': vol_id}, {'id': snap_id})
        self.assertVolumeNames((vol_id,))
        self.assertDictEqual(volumes[vol_name], snapshots[snap_name])
        self.assertIsNot(volumes[vol_name], snapshots[snap_name])

        self.driver.delete_snapshot({'id': snap_id})
        self.assertVolumeNames((vol_id,))
        self.assertDictEqual({}, snapshots)

        self.assertRaisesRegex(exception.VolumeBackendAPIException,
                               'No such snapshot',
                               self.driver.revert_to_snapshot, None,
                               {'id': vol_id}, {'id': snap_id})

        self.driver.delete_volume({'id': vol_id})
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        self.assertRaisesRegex(exception.VolumeBackendAPIException,
                               'No such volume',
                               self.driver.revert_to_snapshot, None,
                               {'id': vol_id}, {'id': snap_id})



# Copyright (C) 2020, 2024, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Utility module for Hitachi HBSD Driver."""

import enum
import functools
import logging as base_logging

from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import units

from cinder import exception
from cinder import utils as cinder_utils
from cinder.volume import volume_types

VERSION = '2.4.0'
CI_WIKI_NAME = 'Hitachi_VSP_CI'
PARAM_PREFIX = 'hitachi'
VENDOR_NAME = 'Hitachi'
DRIVER_DIR_NAME = 'hbsd'
DRIVER_PREFIX = 'HBSD'
DRIVER_FILE_PREFIX = 'hbsd'
TARGET_PREFIX = 'HBSD-'
HDP_VOL_ATTR = 'HDP'
HDT_VOL_ATTR = 'HDT'
NVOL_LDEV_TYPE = 'DP-VOL'
TARGET_IQN_SUFFIX = '.hbsd-target'
PAIR_ATTR = 'HTI'
MIRROR_ATTR = 'GAD'

GIGABYTE_PER_BLOCK_SIZE = units.Gi / 512

PRIMARY_STR = 'primary'
SECONDARY_STR = 'secondary'

NORMAL_LDEV_TYPE = 'Normal'

FULL = 'Full copy'
THIN = 'Thin copy'

INFO_SUFFIX = 'I'
WARNING_SUFFIX = 'W'
ERROR_SUFFIX = 'E'

PORT_ID_LENGTH = 5

BUSY_MESSAGE = "Device or resource is busy."

QOS_KEYS = ['upperIops', 'upperTransferRate',
            'lowerIops', 'lowerTransferRate', 'responsePriority']


@enum.unique
class HBSDMsg(enum.Enum):
    """messages for Hitachi HBSD Driver."""
    DRIVER_INITIALIZATION_START = {
        'msg_id': 4,
        'loglevel': base_logging.INFO,
        'msg': 'Initialization of %(driver)s %(version)s started.',
        'suffix': INFO_SUFFIX,
    }
    SET_CONFIG_VALUE = {
        'msg_id': 5,
        'loglevel': base_logging.INFO,
        'msg': 'Set %(object)s to %(value)s.',
        'suffix': INFO_SUFFIX,
    }
    OBJECT_CREATED = {
        'msg_id': 6,
        'loglevel': base_logging.INFO,
        'msg': 'Created %(object)s. (%(details)s)',
        'suffix': INFO_SUFFIX,
    }
    NO_LUN = {
        'msg_id': 301,
        'loglevel': base_logging.WARNING,
        'msg': 'A LUN (HLUN) was not found. (LDEV: %(ldev)s)',
        'suffix': WARNING_SUFFIX,
    }
    INVALID_LDEV_FOR_UNMAPPING = {
        'msg_id': 302,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to specify a logical device for the volume '
               '%(volume_id)s to be unmapped.',
        'suffix': WARNING_SUFFIX,
    }
    INVALID_LDEV_FOR_DELETION = {
        'msg_id': 304,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to specify a logical device to be deleted. '
               '(method: %(method)s, id: %(id)s)',
        'suffix': WARNING_SUFFIX,
    }
    DELETE_TARGET_FAILED = {
        'msg_id': 306,
        'loglevel': base_logging.WARNING,
        'msg': 'A host group or an iSCSI target could not be deleted. '
               '(port: %(port)s, gid: %(id)s)',
        'suffix': WARNING_SUFFIX,
    }
    CREATE_HOST_GROUP_FAILED = {
        'msg_id': 308,
        'loglevel': base_logging.WARNING,
        'msg': 'A host group could not be added. (port: %(port)s)',
        'suffix': WARNING_SUFFIX,
    }
    CREATE_ISCSI_TARGET_FAILED = {
        'msg_id': 309,
        'loglevel': base_logging.WARNING,
        'msg': 'An iSCSI target could not be added. (port: %(port)s)',
        'suffix': WARNING_SUFFIX,
    }
    UNMAP_LDEV_FAILED = {
        'msg_id': 310,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to unmap a logical device. (LDEV: %(ldev)s)',
        'suffix': WARNING_SUFFIX,
    }
    DELETE_LDEV_FAILED = {
        'msg_id': 313,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to delete a logical device. (LDEV: %(ldev)s)',
        'suffix': WARNING_SUFFIX,
    }
    MAP_LDEV_FAILED = {
        'msg_id': 314,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to map a logical device. (LDEV: %(ldev)s, port: '
               '%(port)s, id: %(id)s, lun: %(lun)s)',
        'suffix': WARNING_SUFFIX,
    }
    DISCARD_ZERO_PAGE_FAILED = {
        'msg_id': 315,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to perform a zero-page reclamation. (LDEV: '
               '%(ldev)s)',
        'suffix': WARNING_SUFFIX,
    }
    ADD_HBA_WWN_FAILED = {
        'msg_id': 317,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to assign the WWN. (port: %(port)s, gid: %(gid)s, '
               'wwn: %(wwn)s)',
        'suffix': WARNING_SUFFIX,
    }
    LDEV_NOT_EXIST = {
        'msg_id': 319,
        'loglevel': base_logging.WARNING,
        'msg': 'The logical device does not exist in the storage system. '
               '(LDEV: %(ldev)s)',
        'suffix': WARNING_SUFFIX,
    }
    REST_LOGIN_FAILED = {
        'msg_id': 321,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to perform user authentication of the REST API server. '
               '(user: %(user)s)',
        'suffix': WARNING_SUFFIX,
    }
    DELETE_PAIR_FAILED = {
        'msg_id': 325,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to delete copy pair. (P-VOL: %(pvol)s, S-VOL: '
               '%(svol)s)',
        'suffix': WARNING_SUFFIX,
    }
    DISCONNECT_VOLUME_FAILED = {
        'msg_id': 329,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to detach the logical device. (LDEV: %(ldev)s, '
               'reason: %(reason)s)',
        'suffix': WARNING_SUFFIX,
    }
    INVALID_EXTRA_SPEC_KEY_PORT = {
        'msg_id': 330,
        'loglevel': base_logging.WARNING,
        'msg': 'The port name specified for the extra spec key '
               '"%(target_ports_param)s" '
               'of the volume type is not specified for the '
               'target_ports or compute_target_ports '
               'parameter in cinder.conf. (port: %(port)s, volume type: '
               '%(volume_type)s)',
        'suffix': WARNING_SUFFIX,
    }
    VOLUME_IS_BEING_REHYDRATED = {
        'msg_id': 333,
        'loglevel': base_logging.WARNING,
        'msg': 'Retyping the volume will be performed using migration '
               'because the specified volume is being rehydrated. '
               'This process may take a long time depending on the data '
               'size. (volume: %(volume_id)s, volume type: %(volume_type)s)',
        'suffix': WARNING_SUFFIX,
    }
    INCONSISTENCY_DEDUPLICATION_SYSTEM_VOLUME = {
        'msg_id': 334,
        'loglevel': base_logging.WARNING,
        'msg': 'Retyping the volume will be performed using migration '
               'because inconsistency was found in the deduplication '
               'system data volume. This process may take a long time '
               'depending on the data size. '
               '(volume: %(volume_id)s, volume type: %(volume_type)s)',
        'suffix': WARNING_SUFFIX,
    }
    HOST_GROUP_NUMBER_IS_MAXIMUM = {
        'msg_id': 335,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to create the host group because the host group '
               'maximum of the port is exceeded. (port: %(port)s)',
        'suffix': WARNING_SUFFIX,
    }
    WWN_NUMBER_IS_MAXIMUM = {
        'msg_id': 336,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to add the wwns to the host group port because the '
               'WWN maximum of the port is exceeded. '
               '(port: %(port)s, WWN: %(wwn)s)',
        'suffix': WARNING_SUFFIX,
    }
    REPLICATION_VOLUME_OPERATION_FAILED = {
        'msg_id': 337,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to %(operation)s the %(type)s in a replication pair. '
               '(volume: %(volume_id)s, reason: %(reason)s)',
        'suffix': WARNING_SUFFIX,
    }
    SITE_INITIALIZATION_FAILED = {
        'msg_id': 338,
        'loglevel': base_logging.WARNING,
        'msg': 'Failed to initialize the driver for the %(site)s storage '
               'system.',
        'suffix': WARNING_SUFFIX,
    }
    INVALID_PORT = {
        'msg_id': 339,
        'loglevel': base_logging.WARNING,
        'msg': 'Port %(port)s will not be used because its settings are '
               'invalid. (%(additional_info)s)',
        'suffix': WARNING_SUFFIX,
    }
    INVALID_PORT_BY_ZONE_MANAGER = {
        'msg_id': 340,
        'loglevel': base_logging.WARNING,
        'msg': 'Port %(port)s will not be used because it is not considered '
               'to be active by the Fibre Channel Zone Manager.',
        'suffix': WARNING_SUFFIX,
    }
    SKIP_DELETING_LDEV = {
        'msg_id': 348,
        'loglevel': base_logging.WARNING,
        'msg': 'Skip deleting the LDEV and its LUNs and pairs because the '
               'LDEV is used by another object. (%(obj)s: %(obj_id)s, LDEV: '
               '%(ldev)s, LDEV label: %(ldev_label)s)',
        'suffix': WARNING_SUFFIX,
    }
    STORAGE_COMMAND_FAILED = {
        'msg_id': 600,
        'loglevel': base_logging.ERROR,
        'msg': 'The command %(cmd)s failed. (ret: %(ret)s, stdout: '
               '%(out)s, stderr: %(err)s)',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_PARAMETER = {
        'msg_id': 601,
        'loglevel': base_logging.ERROR,
        'msg': 'A parameter is invalid. (%(param)s)',
        'suffix': ERROR_SUFFIX,
    }
    PAIR_STATUS_WAIT_TIMEOUT = {
        'msg_id': 611,
        'loglevel': base_logging.ERROR,
        'msg': 'The status change of copy pair could not be '
               'completed. (S-VOL: %(svol)s)',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_STATUS_FOR_COPY = {
        'msg_id': 612,
        'loglevel': base_logging.ERROR,
        'msg': 'The source logical device to be replicated does not exist '
               'in the storage system. (LDEV: %(ldev)s)',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_FOR_EXTENSION = {
        'msg_id': 613,
        'loglevel': base_logging.ERROR,
        'msg': 'The volume %(volume_id)s to be extended was not found.',
        'suffix': ERROR_SUFFIX,
    }
    NO_HBA_WWN_ADDED_TO_HOST_GRP = {
        'msg_id': 614,
        'loglevel': base_logging.ERROR,
        'msg': 'No WWN is assigned. (port: %(port)s, gid: %(gid)s)',
        'suffix': ERROR_SUFFIX,
    }
    UNABLE_TO_DELETE_PAIR = {
        'msg_id': 616,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to delete a pair. (P-VOL: %(pvol)s)',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_VOLUME_TYPE_FOR_EXTEND = {
        'msg_id': 618,
        'loglevel': base_logging.ERROR,
        'msg': 'The volume %(volume_id)s could not be extended. The '
               'volume type must be Normal.',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_FOR_CONNECTION = {
        'msg_id': 619,
        'loglevel': base_logging.ERROR,
        'msg': 'The volume %(volume_id)s to be mapped was not found.',
        'suffix': ERROR_SUFFIX,
    }
    POOL_INFO_RETRIEVAL_FAILED = {
        'msg_id': 620,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to provide information about a pool. (pool: '
               '%(pool)s)',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_FOR_VOLUME_COPY = {
        'msg_id': 624,
        'loglevel': base_logging.ERROR,
        'msg': 'The %(type)s %(id)s source to be replicated was not '
               'found.',
        'suffix': ERROR_SUFFIX,
    }
    CONNECT_VOLUME_FAILED = {
        'msg_id': 634,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to attach the logical device. (LDEV: %(ldev)s, '
               'reason: %(reason)s)',
        'suffix': ERROR_SUFFIX,
    }
    CREATE_LDEV_FAILED = {
        'msg_id': 636,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to add the logical device.',
        'suffix': ERROR_SUFFIX,
    }
    PAIR_TARGET_FAILED = {
        'msg_id': 638,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to add the pair target.',
        'suffix': ERROR_SUFFIX,
    }
    MAP_PAIR_TARGET_FAILED = {
        'msg_id': 639,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to map a logical device to any pair targets. '
               '(LDEV: %(ldev)s)',
        'suffix': ERROR_SUFFIX,
    }
    POOL_NOT_FOUND = {
        'msg_id': 640,
        'loglevel': base_logging.ERROR,
        'msg': 'A pool could not be found. (pool: %(pool)s)',
        'suffix': ERROR_SUFFIX,
    }
    NO_AVAILABLE_RESOURCE = {
        'msg_id': 648,
        'loglevel': base_logging.ERROR,
        'msg': 'There are no resources available for use. (resource: '
               '%(resource)s)',
        'suffix': ERROR_SUFFIX,
    }
    NO_CONNECTED_TARGET = {
        'msg_id': 649,
        'loglevel': base_logging.ERROR,
        'msg': 'The host group or iSCSI target was not found.',
        'suffix': ERROR_SUFFIX,
    }
    RESOURCE_NOT_FOUND = {
        'msg_id': 650,
        'loglevel': base_logging.ERROR,
        'msg': 'The resource %(resource)s was not found.',
        'suffix': ERROR_SUFFIX,
    }
    LDEV_DELETION_WAIT_TIMEOUT = {
        'msg_id': 652,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to delete a logical device. (LDEV: %(ldev)s)',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_ATTR_FOR_MANAGE = {
        'msg_id': 702,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to manage the specified LDEV (%(ldev)s). The LDEV '
               'must be an unpaired %(ldevtype)s.',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_SIZE_FOR_MANAGE = {
        'msg_id': 703,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to manage the specified LDEV (%(ldev)s). The LDEV '
               'size must be expressed in gigabytes.',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_PORT_FOR_MANAGE = {
        'msg_id': 704,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to manage the specified LDEV (%(ldev)s). The LDEV '
               'must not be mapped.',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_TYPE_FOR_UNMANAGE = {
        'msg_id': 706,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to unmanage the volume %(volume_id)s. The volume '
               'type must be %(volume_type)s.',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_LDEV_FOR_MANAGE = {
        'msg_id': 707,
        'loglevel': base_logging.ERROR,
        'msg': 'No valid value is specified for "source-id" or "source-name". '
               'A valid LDEV number must be specified in "source-id" or '
               'a valid LDEV name must be specified in "source-name" '
               'to manage the volume.',
        'suffix': ERROR_SUFFIX,
    }
    FAILED_CREATE_CTG_SNAPSHOT = {
        'msg_id': 712,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to create a consistency group snapshot. '
               'The number of pairs in the consistency group or the number of '
               'consistency group snapshots has reached the limit.',
        'suffix': ERROR_SUFFIX,
    }
    LDEV_NOT_EXIST_FOR_ADD_GROUP = {
        'msg_id': 716,
        'loglevel': base_logging.ERROR,
        'msg': 'No logical device exists in the storage system for the volume '
               '%(volume_id)s to be added to the %(group)s %(group_id)s.',
        'suffix': ERROR_SUFFIX,
    }
    SNAPSHOT_UNMANAGE_FAILED = {
        'msg_id': 722,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to unmanage the snapshot %(snapshot_id)s. '
               'This driver does not support unmanaging snapshots.',
        'suffix': ERROR_SUFFIX,
    }
    INVALID_EXTRA_SPEC_KEY = {
        'msg_id': 723,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to create a volume. '
               'An invalid value is specified for the extra spec key '
               '"%(key)s" of the volume type. (value: %(value)s)',
        'suffix': ERROR_SUFFIX,
    }
    VOLUME_COPY_FAILED = {
        'msg_id': 725,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to copy a volume. (P-VOL: %(pvol)s, S-VOL: %(svol)s)',
        'suffix': ERROR_SUFFIX
    }
    CONSISTENCY_NOT_GUARANTEE = {
        'msg_id': 726,
        'loglevel': base_logging.ERROR,
        'msg': 'A volume or snapshot cannot be deleted. '
               'The consistency of logical device for '
               'a volume or snapshot cannot be guaranteed. (LDEV: %(ldev)s)',
        'suffix': ERROR_SUFFIX
    }
    FAILED_CHANGE_VOLUME_TYPE = {
        'msg_id': 727,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to change a volume type. '
               'An invalid value is specified for the extra spec key '
               '"%(key)s" of the volume type after change. '
               '(value: %(value)s)',
        'suffix': ERROR_SUFFIX
    }
    NOT_COMPLETED_CHANGE_VOLUME_TYPE = {
        'msg_id': 728,
        'loglevel': base_logging.ERROR,
        'msg': 'The volume type change could not be completed. '
               '(LDEV: %(ldev)s)',
        'suffix': ERROR_SUFFIX
    }
    REST_SERVER_CONNECT_FAILED = {
        'msg_id': 731,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to communicate with the REST API server. '
               '(exception: %(exception)s, message: %(message)s, '
               'method: %(method)s, url: %(url)s, params: %(params)s, '
               'body: %(body)s)',
        'suffix': ERROR_SUFFIX,
    }
    REST_API_FAILED = {
        'msg_id': 732,
        'loglevel': base_logging.ERROR,
        'msg': 'The REST API failed. (source: %(errorSource)s, '
               'ID: %(messageId)s, message: %(message)s, cause: %(cause)s, '
               'solution: %(solution)s, code: %(errorCode)s, '
               'method: %(method)s, url: %(url)s, params: %(params)s, '
               'body: %(body)s)',
        'suffix': ERROR_SUFFIX,
    }
    REST_API_TIMEOUT = {
        'msg_id': 733,
        'loglevel': base_logging.ERROR,
        'msg': 'The REST API timed out. (job ID: %(job_id)s, '
               'job status: %(status)s, job state: %(state)s, '
               'method: %(method)s, url: %(url)s, params: %(params)s, '
               'body: %(body)s)',
        'suffix': ERROR_SUFFIX,
    }
    REST_API_HTTP_ERROR = {
        'msg_id': 734,
        'loglevel': base_logging.ERROR,
        'msg': 'The REST API failed. (HTTP status code: %(status_code)s, '
               'response body: %(response_body)s, '
               'method: %(method)s, url: %(url)s, params: %(params)s, '
               'body: %(body)s)',
        'suffix': ERROR_SUFFIX,
    }
    GROUP_OBJECT_DELETE_FAILED = {
        'msg_id': 736,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to delete a %(obj)s in a %(group)s. (%(group)s: '
               '%(group_id)s, %(obj)s: %(obj_id)s, LDEV: %(ldev)s, reason: '
               '%(reason)s)',
        'suffix': ERROR_SUFFIX,
    }
    GROUP_SNAPSHOT_CREATE_FAILED = {
        'msg_id': 737,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to create a volume snapshot in a group snapshot that '
               'does not guarantee consistency. (group: %(group)s, '
               'group snapshot: %(group_snapshot)s, group type: '
               '%(group_type)s, volume: %(volume)s, snapshot: %(snapshot)s)',
        'suffix': ERROR_SUFFIX,
    }
    NO_ACTIVE_WWN = {
        'msg_id': 747,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to initialize volume connection because no active WWN '
               'was found for the connector. (WWN: %(wwn)s, volume: %(volume)s'
               ')',
        'suffix': ERROR_SUFFIX,
    }
    NO_PORT_WITH_ACTIVE_WWN = {
        'msg_id': 748,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to initialize volume connection because no port with '
               'an active WWN was found. (%(port_wwns)s, volume: %(volume)s)',
        'suffix': ERROR_SUFFIX,
    }
    ZONE_MANAGER_IS_NOT_AVAILABLE = {
        'msg_id': 749,
        'loglevel': base_logging.ERROR,
        'msg': 'The Fibre Channel Zone Manager is not available. The Fibre '
               'Channel Zone Manager must be up and running when '
               'port_scheduler parameter is set to True.',
        'suffix': ERROR_SUFFIX,
    }
    HOST_GROUP_OR_WWN_IS_NOT_AVAILABLE = {
        'msg_id': 750,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to initialize volume connection because no available '
               'resource of host group or wwn was found. (ports: %(ports)s)',
        'suffix': ERROR_SUFFIX,
    }
    SITE_NOT_INITIALIZED = {
        'msg_id': 751,
        'loglevel': base_logging.ERROR,
        'msg': 'The driver is not initialized for the %(site)s storage '
               'system.',
        'suffix': ERROR_SUFFIX,
    }
    CREATE_REPLICATION_VOLUME_FAILED = {
        'msg_id': 752,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to create the %(type)s for a %(rep_type)s pair. '
               '(volume: %(volume_id)s, volume type: %(volume_type)s, '
               'size: %(size)s)',
        'suffix': ERROR_SUFFIX,
    }
    DEDUPLICATION_IS_ENABLED = {
        'msg_id': 753,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to create a volume in a %(rep_type)s environment '
               'because deduplication is enabled for the volume type. '
               '(volume: %(volume_id)s, volume type: %(volume_type)s, '
               'size: %(size)s)',
        'suffix': ERROR_SUFFIX,
    }
    CREATE_REPLICATION_PAIR_FAILED = {
        'msg_id': 754,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to create a %(rep_type)s pair or '
               'to mirror data in a %(rep_type)s pair. '
               '(P-VOL: %(pvol)s, S-VOL: %(svol)s, copy group: '
               '%(copy_group)s, pair status: %(status)s)',
        'suffix': ERROR_SUFFIX,
    }
    SPLIT_REPLICATION_PAIR_FAILED = {
        'msg_id': 755,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to split a %(rep_type)s pair. '
               '(P-VOL: %(pvol)s, S-VOL: %(svol)s, '
               'copy group: %(copy_group)s, pair status: %(status)s)',
        'suffix': ERROR_SUFFIX,
    }
    PAIR_CHANGE_TIMEOUT = {
        'msg_id': 756,
        'loglevel': base_logging.ERROR,
        'msg': 'A timeout occurred before the status of '
               'the %(rep_type)s pair changes. '
               '(P-VOL: %(pvol)s, S-VOL: %(svol)s, copy group: '
               '%(copy_group)s, current status: %(current_status)s, '
               'expected status: %(expected_status)s, timeout: %(timeout)s '
               'seconds)',
        'suffix': ERROR_SUFFIX,
    }
    EXTEND_REPLICATION_VOLUME_ERROR = {
        'msg_id': 758,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to extend a volume. The LDEVs for the volume are in '
               'a %(rep_type)s pair and the volume is attached. '
               '(volume: %(volume_id)s, '
               'LDEV: %(ldev)s, source size: %(source_size)s, destination '
               'size: %(destination_size)s, P-VOL: %(pvol)s, S-VOL: %(svol)s, '
               'P-VOL[numOfPorts]: %(pvol_num_of_ports)s, '
               'S-VOL[numOfPorts]: %(svol_num_of_ports)s)',
        'suffix': ERROR_SUFFIX,
    }
    MIGRATE_VOLUME_FAILED = {
        'msg_id': 760,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to migrate a volume. The volume is in a copy pair that '
               'cannot be deleted. (volume: %(volume)s, LDEV: %(ldev)s, '
               '(P-VOL, S-VOL, copy method, status): %(pair_info)s)',
        'suffix': ERROR_SUFFIX,
    }
    REPLICATION_PAIR_ERROR = {
        'msg_id': 766,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to %(operation)s. The LDEV for the volume is in '
               'a remote replication pair. (volume: %(volume)s, '
               '%(snapshot_info)sLDEV: %(ldev)s)',
        'suffix': ERROR_SUFFIX,
    }
    LDEV_NUMBER_NOT_FOUND = {
        'msg_id': 770,
        'loglevel': base_logging.ERROR,
        'msg': 'Failed to %(operation)s. The LDEV number is not found in the '
               'Cinder object. (%(obj)s: %(obj_id)s)',
        'suffix': ERROR_SUFFIX,
    }

    def __init__(self, error_info):
        """Initialize Enum attributes."""
        self.msg_id = error_info['msg_id']
        self.level = error_info['loglevel']
        self.msg = error_info['msg']
        self.suffix = error_info['suffix']

    def output_log(self, storage_id, **kwargs):
        """Output the message to the log file and return the message."""
        msg = self.msg % kwargs
        if storage_id:
            LOG.log(
                self.level,
                "%(storage_id)s MSGID%(msg_id)04d-%(msg_suffix)s: %(msg)s",
                {'storage_id': storage_id[-6:], 'msg_id': self.msg_id,
                 'msg_suffix': self.suffix, 'msg': msg})
        else:
            LOG.log(
                self.level, "MSGID%(msg_id)04d-%(msg_suffix)s: %(msg)s",
                {'msg_id': self.msg_id, 'msg_suffix': self.suffix, 'msg': msg})
        return msg


def output_log(msg_enum, storage_id=None, **kwargs):
    """Output the specified message to the log file and return the message."""
    return msg_enum.output_log(storage_id, **kwargs)


LOG = logging.getLogger(__name__)
MSG = HBSDMsg


def timed_out(start_time, timeout):
    """Check if the specified time has passed."""
    return timeutils.is_older_than(start_time, timeout)


def build_initiator_target_map(connector, target_wwns, lookup_service):
    """Return a dictionary mapping server-wwns and lists of storage-wwns."""
    init_targ_map = {}
    initiator_wwns = connector['wwpns']
    if lookup_service:
        dev_map = lookup_service.get_device_mapping_from_network(
            initiator_wwns, target_wwns)
        for fabric_name in dev_map:
            fabric = dev_map[fabric_name]
            for initiator in fabric['initiator_port_wwn_list']:
                init_targ_map[initiator] = fabric['target_port_wwn_list']
    else:
        for initiator in initiator_wwns:
            init_targ_map[initiator] = target_wwns
    return init_targ_map


def safe_get_err_code(errobj):
    if not errobj:
        return '', ''
    err_code = errobj.get('errorCode', {})
    return err_code.get('SSB1', '').upper(), err_code.get('SSB2', '').upper()


def safe_get_return_code(errobj):
    if not errobj:
        return ''
    err_code = errobj.get('errorCode', {})
    return err_code.get('errorCode', '')


def safe_get_message_id(errobj):
    if not errobj:
        return ''
    return errobj.get('messageId', '')


def safe_get_message(errobj):
    if not errobj:
        return ''
    return errobj.get('message', '')


def is_shared_connection(volume, connector):
    """Check if volume is multiattach to 1 node."""
    connection_count = 0
    host = connector.get('host') if connector else None
    if host and volume.get('multiattach'):
        attachment_list = volume.volume_attachment
        try:
            att_list = attachment_list.object
        except AttributeError:
            att_list = attachment_list
        for attachment in att_list:
            if attachment.attached_host == host:
                connection_count += 1
    return connection_count > 1


def cleanup_cg_in_volume(volume):
    if ('group_id' in volume and volume.group_id and
            'consistencygroup_id' in volume and
            volume.consistencygroup_id):
        volume.consistencygroup_id = None
        if 'consistencygroup' in volume:
            volume.consistencygroup = None


def get_exception_msg(exc):
    if exc.args:
        return exc.msg if isinstance(
            exc, exception.CinderException) else exc.args[0]
    else:
        return ""


def synchronized_on_copy_group():
    def wrap(func):
        @functools.wraps(func)
        def inner(self, remote_client, copy_group_name, *args, **kwargs):
            sync_key = '%s-%s' % (copy_group_name,
                                  self.storage_id[-6:])

            @cinder_utils.synchronized(sync_key, external=True)
            def _inner():
                return func(self, remote_client, copy_group_name,
                            *args, **kwargs)
            return _inner()
        return inner
    return wrap


def get_qos_specs_from_volume(target):
    """Return a dictionary of the QoS specs of the target.

    :param target: Volume or Snapshot whose QoS specs are queried.
    :type target: Volume or Snapshot
    :return: QoS specs.
    :rtype: dict
    """
    # If the target is a Volume, volume_type is volume.volume_type.
    # If the target is a Snapshot, volume_type is snapshot.volume.volume_type.
    # We combine these into "getattr(target, 'volume', target).volume_type)".
    return get_qos_specs_from_volume_type(
        getattr(target, 'volume', target).volume_type)


def get_qos_specs_from_volume_type(volume_type):
    """Return a dictionary of the QoS specs of the volume_type.

    :param volume_type: VolumeType whose QoS specs are queried. This must not
    be None.
    :type volume_type: VolumeType
    :return: QoS specs.
    :rtype: dict
    The following is an example of the returned value:
    {'lowerTransferRate': 7,
     'responsePriority': 2,
     'upperIops': 456}
    """
    qos = {}
    specs = volume_types.get_volume_type_qos_specs(volume_type.id)['qos_specs']
    # The following is an example of the specs:
    # {'consumer': 'back-end',
    #  'created_at': datetime.datetime(2024, 9, 2, 3, 11, 1),
    #  'id': '81058c04-06eb-49d7-9199-7016785bf386',
    #  'name': 'qos1',
    #  'specs': {'lowerTransferRate': '7',
    #            'responsePriority': '2',
    #            'upperIops': '456'}}
    if specs is None:
        return qos
    if 'consumer' in specs and specs['consumer'] not in ('back-end', 'both'):
        return qos
    for key in specs['specs'].keys():
        if key in QOS_KEYS:
            if specs['specs'][key].isdigit():
                qos[key] = int(specs['specs'][key])
            else:
                qos[key] = specs['specs'][key]
    return qos


DICT = '_dict'
CONF = '_conf'


class Config(object):

    def __init__(self, conf):
        super().__setattr__(CONF, conf)
        super().__setattr__(DICT, dict())
        self._opts = {}

    def __getitem__(self, name):
        return (super().__getattribute__(DICT)[name]
                if name in super().__getattribute__(DICT)
                else super().__getattribute__(CONF).safe_get(name))

    def __getattr__(self, name):
        return (super().__getattribute__(DICT)[name]
                if name in super().__getattribute__(DICT)
                else getattr(super().__getattribute__(CONF), name))

    def __setitem__(self, key, value):
        super().__getattribute__(DICT)[key] = value

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def safe_get(self, name):
        return (super().__getattribute__(DICT)[name]
                if name in super().__getattribute__(DICT)
                else super().__getattribute__(CONF).safe_get(name))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Remove quota consistencygroups

Revision ID: b7b88f50aab5
Revises: 9ab1b092a404
Create Date: 2021-11-10 11:54:50.123389
"""

from alembic import op
from sqlalchemy import orm

from cinder.db.sqlalchemy import models


# revision identifiers, used by Alembic.
revision = 'b7b88f50aab5'
down_revision = '9ab1b092a404'
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)

    with session.begin():
        for model in (models.QuotaClass,
                      models.Quota,
                      models.QuotaUsage,
                      models.Reservation):

            session.query(model)\
                .filter_by(deleted=False, resource='consistencygroups')\
                .update(model.delete_values())



# Copyright 2024 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from cinder.scheduler.filters import affinity_filter
from cinder.scheduler.filters import availability_zone_filter
from cinder.scheduler.filters import capabilities_filter
from cinder.scheduler.filters import capacity_filter
from cinder.scheduler.filters import driver_filter
from cinder.scheduler.filters import ignore_attempted_hosts_filter
from cinder.scheduler.filters import instance_locality_filter
from cinder.scheduler.filters import json_filter
from cinder.scheduler.weights import capacity
from cinder.scheduler.weights import chance
from cinder.scheduler.weights import goodness
from cinder.scheduler.weights import stochastic
from cinder.scheduler.weights import volume_number


ALL_FILTER_CLASSES = [
    availability_zone_filter.AvailabilityZoneFilter,
    capabilities_filter.CapabilitiesFilter,
    capacity_filter.CapacityFilter,
    affinity_filter.DifferentBackendFilter,
    driver_filter.DriverFilter,
    instance_locality_filter.InstanceLocalityFilter,
    ignore_attempted_hosts_filter.IgnoreAttemptedHostsFilter,
    json_filter.JsonFilter,
    affinity_filter.SameBackendFilter,
]

ALL_FILTERS = [filter_cls.__name__ for filter_cls in ALL_FILTER_CLASSES]

DEFAULT_SCHEDULER_FILTER_CLASSES = [
    availability_zone_filter.AvailabilityZoneFilter,
    capabilities_filter.CapabilitiesFilter,
    capacity_filter.CapacityFilter,
]

DEFAULT_SCHEDULER_FILTERS = [
    filter_cls.__name__ for filter_cls in DEFAULT_SCHEDULER_FILTER_CLASSES]

ALL_WEIGHER_CLASSES = [
    capacity.AllocatedCapacityWeigher,
    capacity.CapacityWeigher,
    chance.ChanceWeigher,
    goodness.GoodnessWeigher,
    stochastic.StochasticHostWeightHandler,
    volume_number.VolumeNumberWeigher,
]

ALL_WEIGHERS = [weigher.__name__ for weigher in ALL_WEIGHER_CLASSES]
